;; Generated package description from xref.el  -*- no-byte-compile: t -*-
(define-package "xref" "1.4.1" "Cross-referencing commands" '((emacs "26.1")) :commit "890b08e41580a716c2f0b3f1a40ef9b52a19372c" :url "https://elpa.gnu.org/packages/xref.html")
